from factool import Factool

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("factool_blackbox", "response", "claim_info")
class FactoolBlackboxSolver(StandardTaskSolver):
    """
    A solver to process the response using the Factool black box model.
    """
    def __init__(self, args):
        super().__init__(args)
        self.input_prompt = args.get("input_prompt", None)
        self.model_name = self.global_config.get("model_name", "gpt-4o")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        prompt = state.get(self.input_prompt)
        response = state.get(self.input_name)

        factool_instance = Factool(self.model_name)

        inputs = [{"prompt": prompt, "response": response, "category": "kbqa"}]
        claim_info = factool_instance.run(inputs)

        state.set("claim_info", claim_info)
        return True, state



import time
import json
import openai
from typing import List, Dict, Any

from .utils.prompt_base import QGEN_PROMPT
from .utils.api import chatgpt, search_google, search_bing
from .utils.web_util import scrape_url, select_doc_by_keyword_coverage, select_passages_by_semantic_similarity

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("search_engine_evidence_retriever", "claims", "evidences")
class SearchEngineEvidenceRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.search_engine = args.get("search_engine", "google")
        self.search_engine_func = {
            "google": search_google,
            "bing": search_bing
        }.get(self.search_engine, "google")

        self.url_merge_method = args.get("url_merge_method", "union")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)
        queries = self.generate_questions_as_query(claims)
        evidences = self.search_evidence(claims, queries)
        state.set(self.output_name, evidences)
        return True, state

    # generate questions and queries based on a claim
    def generate_questions_as_query(self, claims,
                                    num_retries: int = 3) -> List[list]:
        """
        num_retries: the number of retries when error occurs during openai api calling
        """
        query_list = []
        for i, claim in enumerate(claims):
            for _ in range(num_retries):
                try:
                    response = chatgpt(QGEN_PROMPT + claim)
                    break
                except openai.OpenAIError as exception:
                    print(f"{exception}. Retrying...")
                    time.sleep(1)
            query_list.append(response)
            # print(response)
            # print("\n")

        # convert openai output: a string into a list of questions/queries
        # not check-worthy claims: query response is set as "", accordingly return a []
        # other responses are split into a list of questions/queries
        automatic_query_list = []
        for query in query_list:
            if query == "":
                automatic_query_list.append([])
            else:
                new_tmp = []
                tmp = query.split("\n")
                for q in tmp:
                    q = q.strip()
                    if q == "" or q == "Output:":
                        continue
                    elif q[:6] == "Output":
                        q = q[7:].strip()
                    new_tmp.append(q)
                automatic_query_list.append(new_tmp)

        return automatic_query_list

    # ----------------------------------------------------------
    # Evidence Retrieval
    # ----------------------------------------------------------
    def collect_claim_url_list(self, queries: List[str]) -> List[str]:
        """
        collect urls for a claim given the query list:
        queries: a list of queries or questions for a claim
        search_engine: use which search engine to retrieve evidence, google or bing
        url_union_or_intersection: url operation, to merge all -> 'union' or obtain intersection
        intersection urls tend to be what is not expected, less relevant
        """
        if len(queries) == 0:
            print("Invalid queries: []")
            return None

        urls_list: List[list] = []  # initial list of urls for all queries
        url_query_dict: Dict[str, list] = {}  # url as key, and list of queries corresponding to this url as value.
        url_union, url_intersection = [], []

        for query in queries:
            urls = self.search_engine_func(query)
            urls_list.append(urls)

        for i, urls in enumerate(urls_list):
            for url in urls:
                if url_query_dict.get(url) is None:
                    url_query_dict[url] = [queries[i]]
                else:
                    url_query_dict[url] = url_query_dict[url] + [queries[i]]

        if self.url_merge_method == "union":
            for urls in urls_list:
                url_union += urls
            url_union = list(set(url_union))
            assert (len(url_union) == len(url_query_dict.keys()))
            return list(url_query_dict.keys()), url_query_dict
        elif self.url_merge_method == "intersection":
            url_intersection = urls_list[0]
            for urls in urls_list[1:]:
                url_intersection = list(set(url_intersection).intersection(set(urls)))
            return url_intersection, url_query_dict
        else:
            print("Invalid url operation, please choose from 'union' and 'intersection'.")
            return None, url_query_dict

    def search_evidence(self,
                        decontextualised_claims: List[str],
                        automatic_query_list: List[list],
                        path_save_evidence: str = "evidence.json",
                        save_web_text: bool = False) -> Dict[str, Dict[str, Any]]:

        assert (len(decontextualised_claims) == len(automatic_query_list))

        claim_info: Dict[str, Dict[str, Any]] = {}
        for i, claim in enumerate(decontextualised_claims):
            queries = automatic_query_list[i]
            if len(queries) == 0:
                claim_info[claim] = {"claim": claim, "automatic_queries": queries, "evidence_list": []}
                print("Claim: {} This is an opinion, not check-worthy.".format(claim))
                continue

            # for each checkworthy claim, first gather urls of related web pages
            urls, url_query_dict = self.collect_claim_url_list(queries)

            docs: List[dict] = []
            for j, url in enumerate(urls):
                web_text, _ = scrape_url(url)
                if not web_text is None:
                    docs.append({"query": url_query_dict[url], "url": url, "web_text": web_text})
                else:
                    continue
            print("Claim: {}\nWe retrieved {} urls, {} web pages are accessible.".format(claim, len(urls), len(docs)))

            # we can directly use the first k of url_query_dict, as it is the list of google returned.
            # Here, we select the most relevent top-k docs against the claim by keyword coverage
            # return index of selected documents as the order in docs
            if len(docs) != 0:
                docs_text = [d['web_text'] for d in docs]
                selected_docs_index = select_doc_by_keyword_coverage(claim, docs_text)
                print(selected_docs_index)
            else:
                # no related web articles collected for this claim, continue to next claim
                claim_info[claim] = {"claim": claim, "automatic_queries": queries, "evidence_list": []}
                continue

            selected_docs = [docs_text[i] for i in selected_docs_index]
            # score corresponding passages and select the top-5 passages
            # return the text of passages; and a list of doc ids for each passage.
            # ids here is as the total number and order in selected_docs_index such as in [4, 25, 28, 32, 33]
            topk_passages, passage_doc_id = select_passages_by_semantic_similarity(claim, selected_docs)

            # recover doc_id to original index in docs which records detailed information of a doc
            passage_doc_index = []
            for ids in passage_doc_id:
                passage_doc_index.append([selected_docs_index[id] for id in ids])

            # evidence list
            evidence_list: List[dict] = []
            for pid, p in enumerate(topk_passages):
                doc_ids = passage_doc_index[pid]
                if save_web_text:
                    evidence_list.append({"evidence_id": pid, "web_page_snippet_manual": p,
                                          "query": [docs[doc_id]["query"] for doc_id in doc_ids],
                                          "url": [docs[doc_id]["url"] for doc_id in doc_ids],
                                          "web_text": [docs[doc_id]["web_text"] for doc_id in doc_ids], })
                else:
                    evidence_list.append({"evidence_id": pid, "web_page_snippet_manual": p,
                                          "query": [docs[doc_id]["query"] for doc_id in doc_ids],
                                          "url": [docs[doc_id]["url"] for doc_id in doc_ids],
                                          "web_text": [], })
            claim_info[claim] = {"claim": claim, "automatic_queries": queries, "evidence_list": evidence_list}

        # write to json file
        # Serializing json
        json_object = json.dumps(claim_info, indent=4)

        # Writing to sample.json
        with open(path_save_evidence, "w") as outfile:
            outfile.write(json_object)
        return claim_info



import random
import string

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("ignorant_search_engine_retriever", "claims", "claims_with_evidences")
class IgnorantSearchEngineRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.max_num_documents = args.get("max_num_documents",5)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        claims_with_evidences = {}
        for idx, claim in enumerate(claims):
            # Assume we call some search engine API here
            documents = [string.ascii_letters[random.randint(0, 25)] for i in range(self.max_num_documents)]
            key = f"{idx}-{claim}"  # Convert the tuple to a string key
            claims_with_evidences[key] = documents

        state.set(self.output_name, claims_with_evidences)
        return True, state



import os
import re
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.metrics import classification_report, confusion_matrix

class SnowballingEvaluator():
    """
    Evaluate the LLM responses on the Snowballing dataset.

    Parameters
    ----------
    LLMEvaluator : class
        The LLMEvaluator class.

    Methods
    -------
    evaluate_snowballing(llm_responses: list):
        Evaluate the LLM responses on the Snowballing dataset
    snowballing_barplot(result: dict, fig_path: str, save: bool = False):
        Create a bar plot of the accuracy of the LLM responses on the Snowballing dataset
        for each topic and the overall accuracy.
    get_boolean(response: str, strict=False):
        Get a boolean value from the response.
    """
    def __init__(self):
        pass

    def get_boolean(self, response: str, strict=False):
        """
        Get a boolean value from the response.
        
        """
        low_response = response.lower()
        if strict:
            if low_response.startswith("yes"):
                return True
            elif low_response.startswith("no"):
                return False
            return None
        else:
            # Check if the response contains any of the specified words
            pattern = r"{}".format("|".join(["n't", "no"]))
            if bool(re.search(pattern, response, re.IGNORECASE)):
                return False
            else:
                return True

    def snowballing_barplot(self, result: dict, fig_path: str = "", save: bool = False):
        """
        Create a bar plot of the accuracy of the LLM responses on the Snowballing dataset
        for each topic and the overall accuracy.

        Parameters
        ----------
        cresult : dict
            The evaluation results for the LLM responses on the Snowballing dataset
        fig_path : str
            The path to save the figure.
        save : bool, optional
            Whether to save the figure, by default True.
        """

        # Data
        items = result.keys()

        # Extract the accuracy values for each topic
        values = [round(v["accuracy"], 2) for k, v in result.items()]

        # Create a new figure
        fig, ax = plt.subplots()

        # Plotting
        bars = sns.barplot(x=items, y=values, palette="rocket", hue=items, ax=ax)

        # Adding values on top of each bar
        for bar in bars.patches:
            ax.text(bar.get_x() + bar.get_width() / 2, bar.get_height(), 
                    f'{bar.get_height():.2f}',
                    ha='center', 
                    va='bottom', 
                    color='black', 
                    rotation='horizontal')
            
        # Rotating x-axis tick labels
        plt.xticks(rotation=20) 

        # Set y-axis limits to accommodate annotations
        plt.ylim((0, max(values) + 0.1)) 

        # Adding labels and title
        plt.xlabel("Topics")
        plt.ylabel("Accuracy")
        plt.title("Performance on Snowballing Dataset.")

        if save:
            # Save the figure
            plt.tight_layout()
            plt.savefig(os.path.join(fig_path, "snowballing_barplot.pdf"), format="pdf")
            plt.savefig(os.path.join(fig_path, "snowballing_barplot.png"), format="png")
        
        # Return the figure
        return fig     

    def snowballing_cm(self, labels: list, preds: list, fig_path: str = "", save: bool = False):
        """
        Create a confusion matrix for the Snowballing dataset.

        Parameters
        ----------
        labels : list
            The true labels.
        preds : list
            The predicted labels.
        fig_path : str
            The path to save the figure.
        save : bool, optional
            Whether to save the figure, by default True.
        """

        # Create a new figure
        fig, ax = plt.subplots()

        # Plotting
        cm = sns.heatmap(confusion_matrix(labels, preds), annot=True, fmt="d", cmap="Blues", ax=ax)

        # Adding labels and title
        plt.xticks(ticks=[0.5, 1.5], labels=["True", "False"])
        plt.yticks(ticks=[0.5, 1.5], labels=["True", "False"])
        plt.ylabel("True label")
        plt.xlabel("Predicted label")
        plt.title("Confusion Matrix on Snowballing dataset.")

        if save:
            # Save the figure
            plt.tight_layout()
            plt.savefig(os.path.join(fig_path, "snowballing_cm.pdf"), format="pdf")
            plt.savefig(os.path.join(fig_path, "snowballing_cm.png"), format="png")

        # Return the figure
        return fig       
        
    def evaluate_snowballing(self, llm_responses: list):
        """
        Evaluate the LLM responses on the Snowballing dataset.
        """
        
        # Store evaluation results for three specific topics and aggregate results 
        # for the entire dataset, indexed by topic names.
        results = {} 

        # Define the ground truth answers for the three specific topics.
        topic_answers = {
            "Primality Testing": True,
            "US Senator Search": True,
            "Graph Connectivity-Flight Search": False,
        }

        # Store the responses for each topic.
        topic_responses = {}
        for key in topic_answers:
            topic_responses[key] = []

        # Store the responses for each topic.
        for item in llm_responses:
            topic_responses[item["topic"]].append(self.get_boolean(item["response"]))

        # Evaluate the LLM responses
        labels, preds = [], []
        for key in topic_answers:
            # Evaluate the responses for each topic.
            y_true = [topic_answers[key]] * len(topic_responses[key])
            y_pred = topic_responses[key]
            results[key] = classification_report(y_true, y_pred, output_dict=True, zero_division=0)

            # Aggregate the results for the entire dataset.
            labels += [topic_answers[key]] * len(topic_responses[key])
            preds += topic_responses[key]

        # Evaluate the responses for the entire dataset.
        results["All"] = classification_report(labels, preds, output_dict=True, zero_division=0)

        return results, labels, preds


"""Utils for running the editor."""
import os
import time
from typing import Dict, Union

import openai

openai.api_key = os.getenv("OPENAI_API_KEY")


def parse_api_response(api_response: str) -> str:
    """Extract the agreement gate state and the reasoning from the GPT-3 API response.

    Our prompt returns a reason for the edit and the edit in two consecutive lines.
    Only extract out the edit from the second line.

    Args:
        api_response: Editor response from GPT-3.
    Returns:
        edited_claim: The edited claim.
    """
    api_response = api_response.strip().split("\n")
    if len(api_response) < 2:
        print("Editor error.")
        return None
    edited_claim = api_response[1].split("My fix:")[-1].strip()
    return edited_claim


def run_rarr_editor(
    claim: str,
    query: str,
    evidence: str,
    model: str,
    prompt: str,
    context: str = None,
    num_retries: int = 5,
) -> Dict[str, str]:
    """Runs a GPT-3 editor on the claim given a query and evidence to support the edit.

    Args:
        claim: Text to edit.
        query: Query to guide the editing.
        evidence: Evidence to base the edit on.
        model: Name of the OpenAI GPT-3 model to use.
        prompt: The prompt template to query GPT-3 with.
        num_retries: Number of times to retry OpenAI call in the event of an API failure.
    Returns:
        edited_claim: The edited claim.
    """
    if context:
        gpt3_input = prompt.format(
            context=context, claim=claim, query=query, evidence=evidence
        ).strip()
    else:
        gpt3_input = prompt.format(claim=claim, query=query, evidence=evidence).strip()

    for _ in range(num_retries):
        try:
            response = openai.completions.create(
                model=model,
                prompt=gpt3_input,
                temperature=0.0,
                max_tokens=512,
                stop=["\n\n"],
            )
            break
        except openai.OpenAIError as exception:
            print(f"{exception}. Retrying...")
            time.sleep(2)

    edited_claim = parse_api_response(response.choices[0].text)
    # If there was an error in GPT-3 generation, return the claim.
    if not edited_claim:
        edited_claim = claim
    output = {"text": edited_claim}
    return output



import os
import json
import yaml
from typing import Any
from pathlib import Path
from dotenv import load_dotenv
from collections import namedtuple
from importlib import resources as pkg_resources


from openfactcheck.utils.logging import get_logger, set_verbosity
from openfactcheck.errors import ConfigValidationError
from openfactcheck import templates as solver_config_templates_dir
from openfactcheck import solvers as solver_templates_dir

# Import solver configuration templates
solver_config_templates_path = str(pkg_resources.files(solver_config_templates_dir) / "solver_configs")
solver_config_template_files = [str(f) for f in Path(solver_config_templates_path).iterdir()]


# Import default solvers
# TODO: Currently, only webservice solvers are supported as default solvers
solver_templates_paths = [
    str(pkg_resources.files(solver_templates_dir) / "webservice"),
    str(pkg_resources.files(solver_templates_dir) / "factool"),
    str(pkg_resources.files(solver_templates_dir) / "urdufactcheck"),
]

# Load environment variables from .env file
load_dotenv()


class OpenFactCheckConfig:
    """
    Class to load the OpenFactCheck configuration from a JSON or YAML file.

    Parameters
    ----------
    filename_or_path: str or path object
        The path to the configuration file.

    Attributes
    ----------
    retries: int
        Number of retries for the pipeline components.
    pipeline: namedtuple
        Namedtuple containing the pipeline components.
    solver_configs: dict
        Dictionary containing the solver configurations.
    solver_paths: dict
        Dictionary containing the paths to the solver models.
    output_path: str
        Path to the output directory.
    secrets: namedtuple
        Namedtuple containing the API keys.
    verbose: str
        The verbosity level for the logger.

    Methods
    -------
    solver_configuration(solver: str = None) -> dict:
        Get the solver configuration for a specific solver or all solvers.
    validate():
        Validate the configuration file

    Examples
    --------
    For loading the default configuration file 'config.json':
    >>> config = OpenFactCheckConfig()

    For loading the configuration file from a specific path or filename:
    >>> config = OpenFactCheckConfig("config.json")

    For loading the configuration file and validating it:
    >>> config = OpenFactCheckConfig("config.json")
    >>> config.validate()
    """

    def __init__(self, filename_or_path: str | Path = "config.json"):
        # Setup Logger
        self.logger = get_logger()

        # Set the filename or path
        self.filename_or_path = filename_or_path

        # Define namedtuple structures
        Secrets = namedtuple("Secrets", ["openai_api_key", "serper_api_key", "scraper_api_key"])

        # Define Attributes
        self.config: dict = {}
        self.retries: int = 3
        self.pipeline: list = []
        self.solver_configs: dict[Any, Any] = SolversConfig(solver_config_template_files)()
        self.solver_paths: dict[str, list[str]] = {"default": solver_templates_paths, "user_defined": []}
        self.output_path: str = "tmp/output"
        self.secrets: Secrets = Secrets(openai_api_key=None, serper_api_key=None, scraper_api_key=None)
        self.verbose = "WARNING"

        try:
            # Check if the file exists
            if Path(self.filename_or_path).exists():
                # Loading Config File
                with open(self.filename_or_path, encoding="utf-8") as file:
                    self.config = json.load(file)
                    self.logger.info(f"Config file loaded successfully from {self.filename_or_path}")
            else:
                # Create a dummy configuration file
                self.logger.warning(f"Config file not found: {self.filename_or_path}")

            # Initialize Retries
            if "retries" in self.config:
                self.retries = self.config["retries"]
            else:
                self.logger.warning("No retries found in the configuration file. Using default value of 3.")

            # Initialize template solvers along with the user-defined solvers
            # User defined solvers will override the template solvers
            if "solver_configs" in self.config:
                self.solver_configs = SolversConfig(solver_config_template_files + self.config["solver_configs"])()
            else:
                self.logger.warning(
                    "No solver configurations found in the configuration file. Using default templates only."
                )

            # Initialize template solver paths along with the user-defined solver paths
            if "solver_paths" in self.config:
                self.solver_paths = {
                    "default": solver_templates_paths,
                    "user_defined": self.config["solver_paths"],
                }
            else:
                self.logger.warning(
                    "No solver paths found in the configuration file. Using default solver paths only."
                )

            # Initialize Output Path
            if "output_path" in self.config:
                self.output_path = self.config["output_path"]
                os.makedirs(self.output_path, exist_ok=True)
            else:
                self.logger.warning(
                    "No output path found in the configuration file. Using default output path 'tmp/output'."
                )
                self.output_path = "tmp/output"
                os.makedirs(self.output_path, exist_ok=True)

            # Initialize Pipeline config
            if "pipeline" in self.config:
                self.pipeline = self.config["pipeline"]
            else:
                if self.solver_configs:
                    solvers = list(self.solver_configs.keys())
                    claimprocessor = "factool_claimprocessor" if "factool_claimprocessor" in solvers else None
                    retriever = "factool_retriever" if "factool_retriever" in solvers else None
                    verifier = "factcheckgpt_verifier" if "factcheckgpt_verifier" in solvers else None
                    for solver in solvers:
                        if claimprocessor and retriever and verifier:
                            break
                        if "claimprocessor" in solver:
                            claimprocessor = solver
                        if "retriever" in solver:
                            retriever = solver
                        if "verifier" in solver:
                            verifier = solver
                    self.pipeline = [claimprocessor, retriever, verifier]
                    self.logger.warning(
                        f"No pipeline found in the configuration file. Using first solver as default pipeline. ClaimProcessor: {claimprocessor}, Retriever: {retriever}, Verifier: {verifier}"
                    )

            # Initialize Secrets config
            if "secrets" in self.config:
                self.secrets = Secrets(
                    openai_api_key=self.config["secrets"]["openai_api_key"],
                    serper_api_key=self.config["secrets"]["serper_api_key"],
                    scraper_api_key=self.config["secrets"]["scraper_api_key"],
                )
            else:
                self.logger.warning(
                    "No secrets found in the configuration file. Make sure to set the environment variables."
                )

            # Initialize Environment Variables
            if self.secrets.openai_api_key:
                os.environ["OPENAI_API_KEY"] = self.secrets.openai_api_key
            if self.secrets.serper_api_key:
                os.environ["SERPER_API_KEY"] = self.secrets.serper_api_key
            if self.secrets.scraper_api_key:
                os.environ["SCRAPER_API_KEY"] = self.secrets.scraper_api_key

            # Initialize Verbose
            if "verbose" in self.config:
                self.verbose = self.config["verbose"]
                set_verbosity(self.verbose)
            else:
                self.logger.warning("No verbose level found in the configuration file. Using default level 'WARNING'.")

            # Validate Configuration
            self.validate()

        except FileNotFoundError:
            self.logger.error(f"Config file not found: {self.filename_or_path}")
            raise FileNotFoundError(f"Config file not found: {self.filename_or_path}")

        except json.JSONDecodeError:
            self.logger.error(f"Invalid JSON in config file: {self.filename_or_path}")
            raise ValueError(f"Invalid JSON in config file: {self.filename_or_path}")

        except ConfigValidationError as e:
            self.logger.error(f"Configuration validation failed: {e}")
            raise ConfigValidationError(f"Configuration validation failed: {e}")

        except Exception as e:
            self.logger.error(f"Unexpected error loading config file: {e}")
            raise Exception(f"Unexpected error loading config file: {e}")

    def validate(self):
        """
        Validate the configuration file.

        Raises
        ------
        ValueError
            If the configuration file is invalid.

        Examples
        --------
        >>> config = OpenFactCheckConfig("config.json")
        >>> config.validate()
        """
        # Check for environment variables
        if "OPENAI_API_KEY" not in os.environ:
            self.logger.warning("OPENAI_API_KEY environment variable not found.")
            raise ConfigValidationError("OPENAI_API_KEY environment variable not found.")
        if "SERPER_API_KEY" not in os.environ:
            self.logger.warning("SERPER_API_KEY environment variable not found.")
            raise ConfigValidationError("SERPER_API_KEY environment variable not found.")
        if "SCRAPER_API_KEY" not in os.environ:
            self.logger.warning("SCRAPER_API_KEY environment variable not found.")
            raise ConfigValidationError("SCRAPER_API_KEY environment variable not found.")

    def solver_configuration(self, solver: str | None = None) -> dict:
        """
        Get the solver configuration for a specific solver or all solvers.

        Parameters
        ----------
        solver: str
            The name of the solver to get the configuration for.
            If not provided, returns the configuration for all solvers.

        Returns
        -------
        dict
            The configuration for the specified solver or all solvers.

        Raises
        ------
        ConfigValidationError
            If the configuration validation fails.

        Examples
        --------
        >>> config = OpenFactCheckConfig("config.json")
        >>> config.solver_configuration()
        >>> config.solver_configuration("factcheckgpt_claimprocessor")
        """
        if solver:
            if solver in self.solver_configs:
                return self.solver_configs[solver]
            else:
                self.logger.error(f"Solver not found: {solver}")
                raise ValueError(f"Solver not found: {solver}")
        else:
            return self.solver_configs


class SolversConfig:
    """
    A class to load solver configurations from one or more JSON or YAML files.

    This class reads solver configurations from specified files, merges them,
    and provides access to the combined configuration as a dictionary.

    Parameters
    ----------
    filename_or_paths : str | Path | list[str | Path]
        The path or list of paths to the solver configuration files.

    Attributes
    ----------
    solvers : dict[Any, Any]
        Dictionary containing the merged solver configurations.

    Examples
    --------
    Load solver configurations from a single file:

    >>> solvers = SolversConfig("solvers.yaml")
    >>> config = solvers()

    Load solver configurations from multiple files:

    >>> solvers = SolversConfig(["solvers1.json", "solvers2.yaml"])
    >>> config = solvers()

    Access the solvers dictionary:

    >>> config = solvers()
    """

    def __init__(self, filename_or_path_s: str | Path | list[str] | list[Path]) -> None:
        """
        Initialize the SolversConfig class.

        Parameters
        ----------
        filename_or_path_s: str or path object or list of str or path objects
            The path to the solvers configuration or a list of paths to multiple solvers configurations.
        """
        # Setup Logger
        self.logger = get_logger()

        # Set the filename or path
        self.filename_or_path_or_path_s = filename_or_path_s

        # Define Attributes
        self.solvers: dict[Any, Any] = {}

        try:
            if isinstance(self.filename_or_path_or_path_s, (str, Path)):
                self.__load_config(self.filename_or_path_or_path_s)
            elif isinstance(self.filename_or_path_or_path_s, list):
                self.__load_configs(self.filename_or_path_or_path_s)
            else:
                self.logger.error(f"Invalid filename type: {type(self.filename_or_path_or_path_s)}")
                raise ValueError(f"Invalid filename type: {type(self.filename_or_path_or_path_s)}")

        except FileNotFoundError:
            self.logger.error(f"Solvers file not found: {self.filename_or_path_or_path_s}")
            raise FileNotFoundError(f"Solvers file not found: {self.filename_or_path_or_path_s}")
        except json.JSONDecodeError:
            self.logger.error(f"Invalid JSON in solvers file: {self.filename_or_path_or_path_s}")
            raise ValueError(f"Invalid JSON in solvers file: {self.filename_or_path_or_path_s}")
        except Exception as e:
            self.logger.error(f"Unexpected error loading solvers file: {e}")
            raise Exception(f"Unexpected error loading solvers file: {e}")

    def __load_config(self, filename_or_path: str | Path) -> None:
        # Ensure filename is a string when performing string operations
        filename = str(filename_or_path)

        with open(filename, encoding="utf-8") as file:
            if filename.endswith(".yaml"):
                file_data = yaml.load(file, Loader=yaml.FullLoader)
            elif filename.endswith(".json"):
                file_data = json.load(file)
            else:
                self.logger.error(f"Invalid file format: {filename}")
                raise ValueError(f"Invalid file format: {filename}")

            # Merge current file data into existing solvers dictionary
            self.solvers.update(file_data)

            # Log the loaded configuration pattern
            if "template" in filename:
                self.logger.info(f"Template solver configuration loaded: {filename.split('/')[-1]}")
            else:
                self.logger.info(f"User-defined solver configuration loaded from: {filename}")

    def __load_configs(self, filenames: list[str] | list[Path]) -> None:
        for filename in filenames:
            self.__load_config(filename)

    def __call__(self) -> dict[Any, Any]:
        return self.solvers



import os
import re
import pytz
import datetime
import pandas as pd
from openai import OpenAI
import matplotlib.pyplot as plt

class FreshQAEvaluator():
    def __init__(self):
        self.openai_client = OpenAI()
        self.openai_model_name = "gpt-4o-mini"
        self.current_date = datetime.datetime.now(pytz.timezone("America/Los_Angeles")).strftime("%B %d, %Y")
        self.prefix = (
            "Please evaluate the response to a question under strict evaluation, where"
            " no hallucinations, outdated information, or ill-formed answers are"
            " allowed. Please credit the response only if it provides a confident and"
            " definitive answer, or the correct answer can be obviously inferred from"
            " the response. The primary or final answer when standing alone must be"
            " accurate. Any additional information that is provided must not contradict"
            " the primary answer or reshape one's perception of it. For false-premise"
            " questions, the response must point out the presence of a false premise to"
            " receive credit. For answers that involve names of entities (e.g.,"
            " people), complete names or commonly recognized names are expected."
            " Regarding numerical answers, approximate numbers are generally not"
            " accepted unless explicitly included in the ground-truth answers. A"
            " response that contains any hallucination, no matter how minor, will not"
            " receive credit. Furthermore, when the response indicates that the"
            " information might be outdated, we accept it only if it is evident that"
            " the knowledge has not changed (e.g., through common sense or well-known"
            " facts)."
        )

        self.demo_examples = [
            {
                "question": "How old is the world's oldest verified living person?",
                "correct_answers": [
                    "116 years old",
                    "116",
                ],
                "response": (
                    f"As of today {self.current_date}, the most up-to-date and relevant"
                    " information regarding this query is as follows. The oldest"
                    " verified living person is Maria Branyas Morera, who was born"
                    " on March 4, 1907, making her 116 years old."
                ),
                "comment": (
                    "This is a valid question. The primary answer in the response"
                    " (116 years old) is accurate, and all the provided information"
                    " is accurate and up-to-date. Thus, the response is credited."
                ),
                "evaluation": "correct",
            },
            {
                "question": "When did the UK adopt the Euro?",
                "correct_answers": ["The United Kingdom has never adopted the Euro."],
                "response": (
                    "The UK has never adopted the Euro as its official currency. The"
                    " country has retained the British pound sterling (GBP) as its"
                    " currency."
                ),
                "comment": (
                    "The question contains a false premise. The response debunks the"
                    " question's false premise (i.e., it points out that the UK has"
                    " never adopted the Euro), and all the provided information is"
                    " accurate and up-to-date. Thus, the response is credited."
                ),
                "evaluation": "correct",
            },
            {
                "question": (
                    "How much longer will Brittney Griner spend in the Russian prison?"
                ),
                "correct_answers": [
                    "She was released in December 2022 as part of a prisoner swap."
                ],
                "response": (
                    "I'm sorry, but I have no information to suggest that Brittney"
                    " Griner is currently in a Russian prison. As far as I know, she"
                    " is a professional basketball player who currently plays for the"
                    " Phoenix Mercury in the Women's National Basketball Association"
                    " (WNBA). If you have any more information or context, please let"
                    " me know and I'll do my best to help."
                ),
                "comment": (
                    "The question contains a false premise. The response fails to"
                    " debunk the question's false premise by not explicitly stating"
                    " that Brittney Griner was already released. Additionally, the"
                    " response lacks a confident and definitive answer. Thus, the"
                    " response is not credited."
                ),
                "evaluation": "incorrect",
            },
            {
                "question": (
                    "Which language has the third largest population by number of"
                    " native speakers?"
                ),
                "correct_answers": ["English"],
                "response": "1. Mandarin 2. Spanish 3. English",
                "comment": (
                    "This is a valid question. While the correct answer (English) can"
                    " be inferred from the response, the response lacks an explicit"
                    " statement, which is not desirable."
                    " Thus, the response is not credited."
                ),
                "evaluation": "incorrect",
            },
            {
                "question": (
                    "Has Virginia Woolf's novel about the Ramsay family entered the"
                    " public domain in the United States?"
                ),
                "correct_answers": ["Yes"],
                "response": (
                    "1927's To the Lighthouse is in the public domain in the United"
                    " States. The final answer is To the Lighthouse."
                ),
                "comment": (
                    "This is a valid question. While the primary information in the"
                    " response (1927's To the Lighthouse is in the public domain in"
                    " the United States) is accurate, the final answer when standing"
                    ' alone is inaccurate (it should be "Yes" instead of "To the'
                    ' Lighthouse"). Thus, the response is not credited.'
                ),
                "evaluation": "incorrect",
            },
        ]

        self.demo_questions = [ex["question"] for ex in self.demo_examples]
        self.demo_evaluations = []

        self.demo_evaluation_template = (
            "\ncorrect answer(s): {correct_answers}"
            "\nresponse: {response}"
            "\ncomment: {comment}"
            "\nevaluation: {evaluation}"
        )

        self.evaluation_template = (
            "\ncorrect answer(s): {correct_answers}" "\nresponse: {response}" "\ncomment: "
        )

        self.df = pd.DataFrame()

    def call_openai_api(self, prompt, temperature, max_tokens, chat_completions=True):
        """
        Call the OpenAI API to generate responses.
        """

        # Generate prompts for demo examples
        for ex in self.demo_examples:
            demo_evaluation = self.demo_evaluation_template.format(
                question=ex["question"],
                correct_answers=" | ".join(ex["correct_answers"]),
                response=ex["response"],
                comment=ex["comment"],
                evaluation=ex["evaluation"],
            )
            self.demo_evaluations.append(demo_evaluation)

        # Call the OpenAI API to generate responses
        # If chat completions are enabled, use the chat completions endpoint
        if chat_completions:
            response = self.openai_client.chat.completions.create(
                model=self.openai_model_name,
                temperature=temperature,
                max_tokens=max_tokens,
                messages=[
                    {
                        "role": "system",
                        "content": (
                            "You are a helpful assistant. Respond as concisely as"
                            f" possible. Knowledge cutoff: {self.current_date}."
                        ),
                    },
                    {"role": "user", "content": "What's today's date?"},
                    {
                        "role": "assistant",
                        "content": f"Today is {self.current_date} in Pacific Standard Time.",
                    },
                    {"role": "user", "content": prompt},
                ],
            )
            return response.choices[0].message.content
        
        # If chat completions are disabled, use the completions endpoint
        else:
            response = self.openai_client.completions.create(
                model=self.openai_model_name,
                temperature=temperature,
                max_tokens=max_tokens,
                prompt=prompt,
            )
            return response.choices[0].text


    def call_fresheval(self, prefix, question, response, correct_answers, evaluation):
        """
        Call the FreshEval API to evaluate responses.
        """

        # Set the parameters for the OpenAI API
        temperature = 0.0
        max_tokens = 256
        chat_completions = True

        # Generate prompts for demo examples
        demo_prompts = []
        for q, e in zip(self.demo_questions, self.demo_evaluations):
            demo_prompts.append(f"\n\n\nquestion: {q}{e}")

        # Generate the fresh evaluation prompt
        fresheval_demo = "".join(demo_prompts).strip()
        fresheval_question = f"\n\n\nquestion: {question}{evaluation}"

        # Call the OpenAI API to generate responses
        fresh_eval = prefix + "\n\n\n" + fresheval_demo + fresheval_question
        answer = self.call_openai_api(fresh_eval, temperature, max_tokens, chat_completions)

        return answer

    def extract_ratings(self, response):
        """
        Extract the rating from the evaluation response.
        """

        # If the eval answer contains either of these three words, considered as 0
        # including incorrect, not correct, not credited
        pattern = re.compile(
            r"\b(?:incorrect|not\s+correct|not\s+credited)\b", re.IGNORECASE
        )
        if pattern.search(response):
            return 0
        else:
            return 1
        
    def freshqa_piechart(self, result, fig_path: str = "", save: bool = False):
        """
        Plot a pie chart of the true and false answers on FreshQA.

        Parameters
        ----------
        result : dict
            The evaluation result.
        fig_path : str
            The path to save the figure.
        save : bool, optional
            Whether to save the figure, by default True.
        """

        # Given numbers
        sizes = [result["accuracy"], 1 - result["accuracy"]]
        labels = ["True Answer", "False Answer"]
        colors = [(0, 1, 0, 0.5), (1, 0, 0, 0.5)]  # Red and green with 50% transparency

        # Create a new figure
        fig, ax = plt.subplots()

        # Plot pie chart
        plt.pie(sizes, labels=labels, colors=colors, autopct="%1.1f%%", startangle=140)
        plt.title("Performance on FreshQA Dataset")

        # Equal aspect ratio ensures that pie is drawn as a circle
        plt.axis("equal")

        if save:
            # Save the figure
            plt.tight_layout()
            plt.savefig(os.path.join(fig_path, "freshqa_piechart.pdf"), format="pdf")
            plt.savefig(os.path.join(fig_path, "freshqa_piechart.png"), format="png")

        # Return the figure
        return fig     


    def evaluate_freshqa(self, llm_responses):
        """
        Evaluate the responses generated by the LLM on FreshQA questions.
        """
        
        llm_responses = pd.DataFrame(llm_responses)
        raw_evals = []
        preds = []
        for idx, row in llm_responses.iterrows():
            evaluation = self.evaluation_template.format(
                correct_answers=row["reference_answer"],
                response=row["response"],
            )

            fresheval = self.call_fresheval(
                self.prefix,
                row["question"],
                row["response"],
                row["reference_answer"],
                evaluation,
            )

            evaluation_rating = self.extract_ratings(fresheval)
            raw_evals.append(evaluation)
            preds.append(evaluation_rating)

        # Compute the accuracy (percentage of correct evaluations)
        accuracy = sum(preds) / len(preds)
        result = { "accuracy": accuracy }
        return result, raw_evals, preds




"""All prompts used for RARR prompting."""

QGEN_PROMPT = """I will check things you said and ask questions.

You said: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes. This is to prevent a buildup of mucus. It’s called the nasal cycle.
To verify it,
1. I googled: Does your nose switch between nostrils?
2. I googled: How often does your nostrils switch?
3. I googled: Why does your nostril switch?
4. I googled: What is nasal cycle?

You said: The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford’s psychology building.
To verify it,
1. I googled: Where was Stanford Prison Experiment was conducted?

You said: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
To verify it,
1. I googled: What does Havel-Hakimi algorithm do?
2. I googled: Who are Havel-Hakimi algorithm named after?

You said: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Michael Lloyd.
To verify it,
1. I googled: Who sings the song "Time of My Life"?
2. I googled: Which film is the song "Time of My Life" from?
3. I googled: Who produced the song "Time of My Life"?

You said: Kelvin Hopins was suspended from the Labor Party due to his membership in the Conservative Party.
To verify it,
1. I googled: Why was Kelvin Hopins suspended from Labor Party?

You said: Social work is a profession that is based in the philosophical tradition of humanism. It is an intellectual discipline that has its roots in the 1800s.
To verify it,
1. I googled: What philosophical tradition is social work based on?
2. I googled: What year does social work have its root in?

You said: {claim}
To verify it,
""".strip()

CONTEXTUAL_QGEN_PROMPT = """I will check things you said and ask questions.

Context: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes.
You said: This is to prevent a buildup of mucus. It's called the nasal cycle.
To verify what you just said,
1. I googled: Why does your nostril switch during sleep?
2. I googled: What is nasal cycle?
3. I googled: What is the nostril switching during sleep called?

Context: The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford's psychology building.
You said: It is a psychological study to observe the behaviors of conflict and violence that happen between inmates and prisoners in real prisons.
To verify what you just said,
1. I googled: What type of experiment was the Stanford Prison Experiment?
2. I googled: What was the objective of the Stanford Prison Experiment?

Context: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list.
You said: It is named after Václav Havel and Samih Hakimi.
To verify what you just said,
1. I googled: Who are Havel-Hakimi algorithm named after?

Context: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing.
You said: The song was produced by Michael Lloyd in the same year.
To verify what you just said,
1. I googled: Who produced the song "Time of My Life"?
2. I googled: When was the song "Time of My Life" by Bill Medley produced?

Context: The Late Show with Stephen Colbert is an American late-night talk show hosted by Stephen Colbert, which premiered on September 8, 2015.
You said: Produced by Spartina Productions and CBS Television Studios, it is the second iteration of CBS' Late Show franchise.
To verify what you just said,
1. I googled: Who produces "The Late Show with Stephen Colbert"?
2. I googled: What are the iterations of CBS' Late Show franchise?

Context: Super Mario Sunshine was released on GameCube in 2002. In the game, Mario uses a tool strapped to his back called FLUDD, which stands for The Flash Liquidizer Ultra Dousing Device.
You said: It can be used to spray water at objects or enemies. This allows Mario to change his movements, kill enemies, or clean up hazards on the floor.
To verify what you just said,
1. I googled: What is the main function of FLUDD in Super Mario Sunshine?
2. I googled: What can FLUDD in Super Mario Sunshine be used on?

Context: {context}
You said: {claim}
To verify what you just said,
""".strip()

AGREEMENT_GATE_PROMPT = """I will check some things you said.

1. You said: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes. This is to prevent a buildup of mucus. It’s called the nasal cycle.
2. I checked: How often do your nostrils switch?
3. I found this article: Although we don’t usually notice it, during the nasal cycle one nostril becomes congested and thus contributes less to airflow, while the other becomes decongested. On average, the congestion pattern switches about every 2 hours, according to a small 2016 study published in the journal PLOS One.
4. Reasoning: The article said the nose’s switching time is about every 2 hours, and you said the nose's switching time is about every 45 minutes.
5. Therefore: This disagrees with what you said.

1. You said: The Little House books were written by Laura Ingalls Wilder. The books were published by HarperCollins.
2. I checked: Who published the Little House books?
3. I found this article: These are the books that started it all -- the stories that captured the hearts and imaginations of children and young adults worldwide. Written by Laura Ingalls Wilder and published by HarperCollins, these beloved books remain a favorite to this day.
4. Reasoning: The article said the Little House books were published by HarperCollins and you said the books were published by HarperCollins.
5. Therefore: This agrees with what you said.

1. You said: Real Chance of Love was an American reality TV show. Season 2 of the show was won by Cali, who chose to be with Chance.
2. I checked: Who won season 2 of Real Chance of Love?
3. I found this article: Real Chance of Love 2: Back in the Saddle is the second season of the VH1 reality television dating series Real Chance of Love. Ahmad Givens (Real) and Kamal Givens (Chance), former contestants on I Love New York are the central figures.
4. Reasoning: The article doesn't answer the question and you said that Cali won season 2 of Real Chance of Love.
5. Therefore: This is irrelevant to what you said.

1. You said: The Stanford Prison Experiment was conducted in the basement of Jordan Hall, Stanford’s psychology building.
2. I checked: Where was Stanford Prison Experiment conducted?
3. I found this article: Carried out August 15-21, 1971 in the basement of Jordan Hall, the Stanford Prison Experiment set out to examine the psychological effects of authority and powerlessness in a prison environment.
4. Reasoning: The article said the Stanford Prison Experiment was conducted in Jordan Hall and you said the Stanford Prison Experiment was conducted in Jordan Hall.
5. Therefore: This agrees with what you said.

1. You said: Social work is a profession that is based in the philosophical tradition of humanism. It is an intellectual discipline that has its roots in the 1800s.
2. I checked: When did social work have its roots?
3. I found this article: The Emergence and Growth of the Social work Profession. Social work’s roots were planted in the 1880s, when charity organization societies (COS) were created to organize municipal voluntary relief associations and settlement houses were established.
4. Reasoning: The article said social work has its roots planted in the 1880s and you said social work has its root in the 1800s.
5. Therefore: This disagrees with what you said.

1. You said: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
2. I checked: What is the Havel-Hakimi algorithm?
3. I found this article: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. This construction is based on a recursive algorithm. The algorithm was published by Havel (1955), and later by Hakimi (1962).
4. Reasoning: The article said the Havel-Hakimi algorithm is for constructing a special solution if a simple graph for the given degree sequence exists and you said the Havel-Hakimi algorithm is for converting the adjacency matrix of a graph.
5. Therefore: This disagrees with what you said.

1. You said: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Michael Lloyd.
2. I checked: Who was the producer of "(I’ve Had) The Time of My Life"?
3. I found this article: On September 8, 2010, the original demo of this song, along with a remix by producer Michael Lloyd , was released as digital files in an effort to raise money for the Patrick Swayze Pancreas Cancer Resarch Foundation at Stanford University.
4. Reasoning: The article said that a demo was produced by Michael Lloyd and you said "Time of My Life" was produced by Michael Lloyd.
5. Therefore: This agrees with what you said.

1. You said: Tiger Woods is the only player who has won the most green jackets. He has won four times. The Green Jacket is one of the most coveted prizes in all of golf.
2. I checked: What is the Green Jacket in golf?
3. I found this article: The green jacket is a classic, three-button, single-breasted and single-vent, featuring the Augusta National Golf Club logo on the left chest pocket. The logo also appears on the brass buttons.
4. Reasoning: The article said the Green Jacket is a classic three-button single-breasted and single-vent and you said the Green Jacket is one of the most coveted prizes in all of golf.
5. Therefore: This is irrelevant to what you said.

1. You said: Kelvin Hopins was suspended from the Labor Party because he had allegedly sexually harassed and behaved inappropriately towards a Labour Party activist, Ava Etemadzadeh.
2. I checked: Why was Kelvin Hopins suspeneded from the Labor Party?
3. I found this article: A former Labour MP has left the party before an inquiry into sexual harassment allegations against him was able to be concluded, the party has confirmed. Kelvin Hopkins was accused in 2017 of inappropriate physical contact and was suspended by the Labour party pending an investigation.
4. Reasoning: The article said Kelvin Hopins was suspended because of inappropriate physical contact and you said that Kelvin Hopins was suspended because he allegedly sexually harassed Ava Etemadzadeh.
5. Therefore: This agrees with what you said.

1. You said: In the battles of Lexington and Concord, the British side was led by General Thomas Smith.
2. I checked: Who led the British side in the battle of Lexington and Concord?
3. I found this article: Interesting Facts about the Battles of Lexington and Concord. The British were led by Lieutenant Colonel Francis Smith. There were 700 British regulars.
4. Reasoning: The article said the British side was led by Lieutenant Colonel Francis Smith and you said the British side was led by General Thomas Smith.
5. Therefore: This disagrees with what you said.

1. You said: {claim}
2. I checked: {query}
3. I found this article: {evidence}
4. Reasoning:
""".strip()

CONTEXTUAL_AGREEMENT_GATE_PROMPT = """I will check some things you said.

1. Context: Your nose switches back and forth between nostrils. It's called the nasal cycle. This is to prevent a buildup of mucus.
2. You said: When you sleep, you switch about every 45 minutes.
3. I checked: How often do your nostrils switch?
4. I found this article: Although we don’t usually notice it, during the nasal cycle one nostril becomes congested and thus contributes less to airflow, while the other becomes decongested. On average, the congestion pattern switches about every 2 hours, according to a small 2016 study published in the journal PLOS One.
5. Reasoning: The article said the nose’s switching time is about every 2 hours, and you said the nose's switching time is about every 45 minutes.
6. Therefore: This disagrees with what you said.

1. Context: The Little House books is a series of American children's novels.
2. You said: The books were published by HarperCollins.
3. I checked: Who published the Little House books?
4. I found this article: These are the books that started it all -- the stories that captured the hearts and imaginations of children and young adults orldwide. Written by Laura Ingalls Wilder and published by HarperCollins, these beloved books remain a favorite to this day.
5. Reasoning: The article said the Little House books were published by HarperCollins and you said the books were published by HarperCollins.
6. Therefore: This agrees with what you said.

1. Context: Real Chance of Love was an American reality TV show.
2. You said: Season 2 of the show was won by Cali, who chose to be with Chance.
3. I checked: Who won season 2 of Real Chance of Love?
4. I found this article: Real Chance of Love 2: Back in the Saddle is the second season of the VH1 reality television dating series Real Chance of Love. Ahmad Givens (Real) and Kamal Givens (Chance), former contestants on I Love New York are the central figures.
5. Reasoning: The article doesn't answer the question and you said that Cali won season 2 of Real Chance of Love.
6. Therefore: This is irrelevant to what you said.

1. Context: The Stanford Prison Experiment is a psychological study to observe the behaviors of conflict and violence that happen between inmates and prisoners in real prisons.
2. You said: It was conducted in the basement of Jordan Hall, Stanford’s psychology building.
3. I checked: Where was Stanford Prison Experiment conducted?
4. I found this article: Carried out August 15-21, 1971 in the basement of Jordan Hall, the Stanford Prison Experiment set out to examine the psychological effects of authority and powerlessness in a prison environment.
5. Reasoning: The article said the Stanford Prison Experiment was conducted in Jordan Hall and you said the Stanford Prison Experiment was conducted in Jordan Hall.
6. Therefore: This agrees with what you said.

1. Context: Social work is a profession that is based in the philosophical tradition of humanism.
2. You said: It is an intellectual discipline that has its roots in the 1800s.
3. I checked: When did social work have its roots?
4. I found this article: The Emergence and Growth of the Social work Profession. Social work’s roots were planted in the 1880s, when charity organization societies (COS) were created to organize municipal voluntary relief associations and settlement houses were established.
5. Reasoning: The article said social work has its roots planted in the 1880s and you said social work has its root in the 1800s.
6. Therefore: This disagrees with what you said.

1. Context: The Havel-Hakimi algorithm is named after Václav Havel and Samih Hakimi.
2. You said: It is an algorithm for converting the adjacency matrix of a graph into its adjacency list.
3. I checked: What is the Havel-Hakimi algorithm?
4. I found this article: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. This construction is based on a recursive algorithm. The algorithm was published by Havel (1955), and later by Hakimi (1962).
5. Reasoning: The article said the Havel-Hakimi algorithm is for constructing a special solution if a simple graph for the given degree sequence exists and you said the Havel-Hakimi algorithm is for converting the adjacency matrix of a graph.
6. Therefore: This disagrees with what you said.

1. Context: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing.
2. You said: The song was produced by Michael Lloyd.
3. I checked: Who was the producer of "(I’ve Had) The Time of My Life"?
4. I found this article: On September 8, 2010, the original demo of this song, along with a remix by producer Michael Lloyd , was released as digital files in an effort to raise money for the Patrick Swayze Pancreas Cancer Resarch Foundation at Stanford University.
5. Reasoning: The article said that a demo was produced by Michael Lloyd and you said "Time of My Life" was produced by Michael Lloyd.
6. Therefore: This agrees with what you said.

1. Context: Tiger Woods is the only player who has won the most green jackets. He has won four times.
2. You said: The Green Jacket is one of the most coveted prizes in all of golf.
3. I checked: What is the Green Jacket in golf?
4. I found this article: The green jacket is a classic, three-button, single-breasted and single-vent, featuring the Augusta National Golf Club logo on the left chest pocket. The logo also appears on the brass buttons.
5. Reasoning: The article said the Green Jacket is a classic three-button single-breasted and single-vent and you said the Green Jacket is one of the most coveted prizes in all of golf.
6. Therefore: This is irrelevant to what you said.

1. Context: Kelvin Hopins was suspended from the Labor Party.
2. You said: This was because he had allegedly sexually harassed and behaved inappropriately towards a Labour Party activist, Ava Etemadzadeh.
3. I checked: Why was Kelvin Hopins suspeneded from the Labor Party?
4. I found this article: A former Labour MP has left the party before an inquiry into sexual harassment allegations against him was able to be concluded, the party has confirmed. Kelvin Hopkins was accused in 2017 of inappropriate physical contact and was suspended by the Labour party pending an investigation.
5. Reasoning: The article said Kelvin Hopins was suspended because of inappropriate physical contact and you said that Kelvin Hopins was suspended because he allegedly sexually harassed Ava Etemadzadeh.
6. Therefore: This agrees with what you said.

1. Context: The Battles of Lexington and Concord, fought on April 19, 1775, kicked off the American Revolutionary War (1775-83).
2. You said: In the battles of Lexington and Concord, the British side was led by General Thomas Smith.
3. I checked: Who led the British side in the battle of Lexington and Concord?
4. I found this article: Interesting Facts about the Battles of Lexington and Concord. The British were led by Lieutenant Colonel Francis Smith. There were 700 British regulars.
5. Reasoning: The article said the British side was led by Lieutenant Colonel Francis Smith and you said the British side was led by General Thomas Smith.
6. Therefore: This disagrees with what you said.

1. Context: {context}
2. You said: {claim}
3. I checked: {query}
4. I found this article: {evidence}
5. Reasoning:
""".strip()

EDITOR_PROMPT = """I will fix some things you said.

1. You said: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes. This is to prevent a buildup of mucus. It’s called the nasal cycle.
2. I checked: How often do your nostrils switch?
3. I found this article: Although we don’t usually notice it, during the nasal cycle one nostril becomes congested and thus contributes less to airflow, while the other becomes decongested. On average, the congestion pattern switches about every 2 hours, according to a small 2016 study published in the journal PLOS One.
4. This suggests 45 minutes switch time in your statement is wrong.
5. My fix: Your nose switches back and forth between nostrils. When you sleep, you switch about every 2 hours. This is to prevent a buildup of mucus. It’s called the nasal cycle.

1. You said: In the battles of Lexington and Concord, the British side was led by General Thomas Hall.
2. I checked: Who led the British side in the battle of Lexington and Concord?
3. I found this article: Interesting Facts about the Battles of Lexington and Concord. The British were led by Lieutenant Colonel Francis Smith. There were 700 British regulars.
4. This suggests General Thomas Hall in your statement is wrong.
5. My fix: In the battles of Lexington and Concord, the British side was led by Lieutenant Colonel Francis Smith.

1. You said: The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford’s psychology building.
2. I checked: Where was Stanford Prison Experiment conducted?
3. I found this article: Carried out August 15-21, 1971 in the basement of Jordan Hall, the Stanford Prison Experiment set out to examine the psychological effects of authority and powerlessness in a prison environment.
4. This suggests Encina Hall in your statement is wrong.
5. My fix: The Stanford Prison Experiment was conducted in the basement of Jordan Hall, Stanford’s psychology building.

1. You said: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
2. I checked: What is the Havel-Hakimi algorithm?
3. I found this article: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. This construction is based on a recursive algorithm. The algorithm was published by Havel (1955), and later by Hakimi (1962).
4. This suggests the Havel-Hakimi algorithm’s functionality in your statement is wrong.
5. My fix: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. It is named after Vaclav Havel and Samih Hakimi.

1. You said: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Phil Ramone.
2. I checked: Who was the producer of "(I’ve Had) The Time of My Life"?
3. I found this article: On September 8, 2010, the original demo of this song, along with a remix by producer Michael Lloyd , was released as digital files in an effort to raise money for the Patrick Swayze Pancreas Cancer Resarch Foundation at Stanford University.
4. This suggests "Time of My Life" producer name in your statement is wrong.
5. My fix: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Michael Lloyd.

1. You said: Phoenix Market City Pune is located on 21 acres of prime property in Pune. It is spread across four levels with approximately 1.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.
2. I checked: What is the area of Phoenix Market City in Pune?
3. I found this article: Phoenix Market City was opened in January 2013 and has the distinction of being the largest mall in the city of Pune, with the area of 3.4 million square feet. It is located in the Viman Nagar area of Pune.
4. This suggests the 1.4 million square feet of built-up space in your statment is wrong.
5. My fix: Phoenix Market City Pune is located on 21 acres of prime property in Pune. It is spread across four levels with approximately 3.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.

1. You said: {claim}
2. I checked: {query}
3. I found this article: {evidence}
4. This suggests
""".strip()

CONTEXTUAL_EDITOR_PROMPT = """I will fix some things you said.

1. Context: Your nose switches back and forth between nostrils. It's called the nasal cycle. This is to prevent a buildup of mucus.
2. You said: When you sleep, you switch about every 45 minutes.
3. I checked: How often do your nostrils switch?
4. I found this article: Although we don’t usually notice it, during the nasal cycle one nostril becomes congested and thus contributes less to airflow, while the other becomes decongested. On average, the congestion pattern switches about every 2 hours, according to a small 2016 study published in the journal PLOS One.
5. This suggests 45 minutes switch time in your statement is wrong.
6. My fix: When you sleep, you switch about every 2 hours.

1. Context: The Battles of Lexington and Concord, fought on April 19, 1775, kicked off the American Revolutionary War (1775-83).
2. You said: In the battles of Lexington and Concord, the British side was led by General Thomas Hall.
3. I checked: Who led the British side in the battle of Lexington and Concord?
4. I found this article: Interesting Facts about the Battles of Lexington and Concord. The British were led by Lieutenant Colonel Francis Smith. There were 700 British regulars.
5. This suggests General Thomas Hall in your statement is wrong.
6. My fix: In the battles of Lexington and Concord, the British side was led by Lieutenant Colonel Francis Smith.

1. Context: The Stanford Prison Experiment is a psychological study to observe the behaviors of conflict and violence that happen between inmates and prisoners in real prisons.
2. You said: It was conducted in the basement of Encina Hall, Stanford’s psychology building.
3. I checked: Where was Stanford Prison Experiment conducted?
4. I found this article: Carried out August 15-21, 1971 in the basement of Jordan Hall, the Stanford Prison Experiment set out to examine the psychological effects of authority and powerlessness in a prison environment.
5. This suggests Encina Hall in your statement is wrong.
6. My fix: It was conducted in the basement of Jordan Hall, Stanford’s psychology building.

1. Context: The Havel-Hakimi algorithm is named after Václav Havel and Samih Hakimi.
2. You said: It is an algorithm for converting the adjacency matrix of a graph into its adjacency list.
3.. I checked: What is the Havel-Hakimi algorithm?
4. I found this article: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. This construction is based on a recursive algorithm. The algorithm was published by Havel (1955), and later by Hakimi (1962).
5. This suggests the Havel-Hakimi algorithm’s functionality in your statement is wrong.
6. My fix: It is an algorithm for constructing a special solution if a simple graph for the given degree sequence exists, or proving that one cannot find a positive answer.

1. Context: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing.
2. You said: The song was produced by Phil Ramone.
3. I checked: Who was the producer of "(I’ve Had) The Time of My Life"?
4. I found this article: On September 8, 2010, the original demo of this song, along with a remix by producer Michael Lloyd , was released as digital files in an effort to raise money for the Patrick Swayze Pancreas Cancer Resarch Foundation at Stanford University.
5. This suggests "Time of My Life" producer name in your statement is wrong.
6. My fix: The song was produced by Michael Lloyd.

1. Context: Phoenix Market City Pune is located on 21 acres of prime property in Pune.
2. You said: Phoenix Market City is spread across four levels with approximately 1.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.
3. I checked: What is the area of Phoenix Market City in Pune?
4. I found this article: Phoenix Market City was opened in January 2013 and has the distinction of being the largest mall in the city of Pune, with the area of 3.4 million square feet. It is located in the Viman Nagar area of Pune.
5. This suggests the 1.4 million square feet of built-up space in your statment is wrong.
6. My fix: Phoenix Market City is spread across four levels with approximately 3.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.

1. Context: {context}
2. You said: {claim}
3. I checked: {query}
4. I found this article: {evidence}
5. This suggests
""".strip()



from .rarr_utils.question_generation import run_rarr_question_generation
from .prompts import rarr_prompts

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver


@Solver.register("rarr_question_generator", "claims_with_context", "claims_with_questions")
class RARRQuestionGenerator(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("model", "gpt-4o-instruct")
        self.temperature_qgen = args.get("temperature_qgen", 0.7)
        self.num_rounds_qgen = args.get("num_rounds_qgen", 3)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)
        # should be DICT[Str, DICT[Str, Any]]
        if type(claims) == list:
            claims = {c: dict() for c in claims}
        for claim, contents in claims.items():
            context = contents.get("context", None)
            claims[claim]["questions"] = run_rarr_question_generation(
                claim=claim,
                context=context,
                model=self.model,
                prompt=rarr_prompts.CONTEXTUAL_QGEN_PROMPT if context else rarr_prompts.QGEN_PROMPT,
                temperature=self.temperature_qgen,
                num_rounds=self.num_rounds_qgen,
            )

        state.set(self.output_name, claims)
        return True, state



import os
import streamlit as st

@st.dialog("Secrets Configuration")
def get_secrets():
    st.markdown("Please provide the following keys to evaluate LLM response:")
    st.info("Keys defined in your environment variables are automatically loaded.")

    # OpenAI API Key
    st.markdown("You can find the OpenAI API Key [here](https://platform.openai.com/account/api-keys).")
    openai_api_key = st.text_input("OpenAI API Key", type="password", value=os.getenv("OPENAI_API_KEY", default=""))
    os.environ["OPENAI_API_KEY"] = openai_api_key

    # Serper API Key
    st.markdown("You can find the Serper API Key [here](https://serpapi.com/dashboard).")
    serper_api_key = st.text_input("Serper API Key", type="password", value=os.getenv("SERPER_API_KEY", default=""))
    os.environ["SERPER_API_KEY"] = serper_api_key

    # ScraperAPI Key
    st.markdown("You can find the ScraperAPI Key [here](https://www.scraperapi.com/).")
    scraper_api_key = st.text_input("ScraperAPI Key", type="password", value=os.getenv("SCRAPER_API_KEY", default=""))
    os.environ["SCRAPER_API_KEY"] = scraper_api_key

    if st.button("Submit"):
        st.session_state.api_keys = True
        st.rerun()



from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("useless_response_regenerator", "claims_with_tags", "output")
class UselessResponseRegenerator(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        true_claims = [k[1] for k, v in claims.items() if v is True]
        new_response = ' '.join(true_claims)
        state.set(self.output_name, new_response)
        return True, state



import json

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("factool_blackbox_post_editor", "claim_info", "claim_info")
class FactoolBlackboxPostEditor(StandardTaskSolver):
    """
    A solver to post-process the results of the Factool black box model.
    Used to presents the results in human-readable format and to save the analysis in a JSON file.
    """
  
    def __init__(self, args):
        super().__init__(args)
        self.path_save_analysis = args.get("path_save_analysis","factool_evidence_analysis.json")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)

        # Restructure some of the output for concatenation (corrected claims)
        edited_claims = ''
        for clf in claim_info['detailed_information'][0]['claim_level_factuality']:
            edited_claims += 'Claim: "' + clf['claim'] + '" => '
            edited_claims += ('' if (clf['error'] == 'None' or len(clf['error']) == 0) else (clf['error'] + ' '))
            edited_claims += ('' if (clf['reasoning'] == 'None' or len(clf['reasoning']) == 0) else clf['reasoning'])
            edited_claims += ((' ' + clf['claim']) if (clf['correction'] == 'None' or len(clf['correction']) == 0) else (' ' + clf['correction']))
            edited_claims += '\n'
        edited_claims = edited_claims[:-1]
        new_claim_info = {}
        new_claim_info[claim_info['detailed_information'][0]['response']] = {
            "edited_claims": edited_claims
        }

        # Serializing json
        json_object = json.dumps(claim_info, indent=4)

        # Writing to sample.json
        with open(self.path_save_analysis, "w") as outfile:
            outfile.write(json_object)

        state.set(self.output_name, new_claim_info)
        return True, state
